///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////

/*
 * \file IncrementalWriter.h
 * \author tsharpe
 * \date Aug 18, 2009
 *
 * \brief
 */

#ifndef FEUDAL_INCREMENTALWRITER_H_
#define FEUDAL_INCREMENTALWRITER_H_

#include "feudal/FeudalFileWriter.h"

/// A class for writing feudal files for some particular feudal type.
/// The template parameter T must be a class that has a value_type, the
/// method char* T::writeFeudal( BinaryWriter& ), and the static method
/// T::fixedDataLen().
template <class T>
class IncrementalWriter
{
public:
    IncrementalWriter( char const* fileName, unsigned long estimatedSize = 1000000 )
    : mWriter(fileName,sizeof(T),sizeof(typename T::value_type),
              T::fixedDataLen(),estimatedSize)
    {}

    // compiler-supplied destructor is OK

    void add( T const& val )
    { unsigned long buf;
      mWriter.addElement(val.writeFeudal(mWriter.getWriter(),&buf)); }

    template <class Itr>
    void add( Itr begin, Itr end )
    { while ( begin != end ) { add(*begin); ++begin; } }

    void checkPoint() { mWriter.checkPoint(); }

    unsigned long getNElements() { return mWriter.getNElements(); }

    void close() { mWriter.close(); }

private:
    IncrementalWriter( IncrementalWriter const& ); // unimplemented -- no copying
    IncrementalWriter& operator=( IncrementalWriter const& ); // unimplemented -- no copying

    FeudalFileWriter mWriter;
};

#endif /* FEUDAL_INCREMENTALWRITER_H_ */
