////////////////////////////////////////////////////////////////////////////
//                  SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//      This software and its documentation are copyright (2009) by the   //
//  Broad Institute.  All rights are reserved.  This software is supplied //
//  without any warranty or guaranteed support whatsoever. The Broad      //
//  Institute is not responsible for its use, misuse, or functionality.   //
////////////////////////////////////////////////////////////////////////////
/*
 * \file QualNibbleVec.cc
 * \author tsharpe
 * \date Oct 21, 2009
 *
 * \brief
 */
#ifndef FORCE_DEBUG
#define NDEBUG
#endif

#include "feudal/QualNibbleVec.h"
#include "feudal/FieldVecDefs.h"
#include "feudal/OuterVecDefs.h"

template class FieldVec< 4, MempoolAllocator<unsigned char> >;
template class OuterVec<QualNibbleVec>;

void WriteAll( VecQualNibbleVec const& quals, String const& outfile )
{
    IncrementalWriter<qualvector> writer(outfile.c_str());
    for ( uint64_t i = 0; i < quals.size(); ++i )
        writer.add(quals[i].GetQualvector());
    writer.close();
}




void LoadQualNibbleVec( const String & quals_file,
			VecQualNibbleVec * quals,
			const bool CAP_QUALS,
			const size_t QUAL_CEIL_RADIUS )
{
  // Load quality scores.
  quals->clear();

  typedef VirtualMasterVec<qualvector> FvvQv;

  //cout << Tag() << "Loading quality scores from " << quals_file << endl;
  FvvQv quals_full( quals_file.c_str( ) );
  for (FvvQv::const_iterator qItr = quals_full.begin();
       qItr != quals_full.end(); qItr++)
    quals->push_back(QualNibbleVec(*qItr));

  // ------------------
  // Cap quality scores
  // ------------------
  if ( CAP_QUALS && QUAL_CEIL_RADIUS > 0 ) {
    //cout << Tag() << "Capping quality scores" << endl;
    CapQualityScores(QUAL_CEIL_RADIUS, quals);
  }
}

// Helper function for LoadQualNibbleVec.
template<class Q_t>
void CapQualityScores(const int qual_ceil_radius, Q_t * quals) 
{
      vec<int> newvals;
      size_t num_reads = quals->size();
      for ( size_t id = 0; id < num_reads; id++ )
        {    newvals.resize_and_set( (*quals)[id].size( ), 1000000 );
        for ( longlong j = 0; j < (*quals)[id].size( ); j++ )
          {    for ( longlong u = Max( (longlong) 0, j - qual_ceil_radius );
                     u <= Min( (longlong) (*quals)[id].size( ) - 1, 
                               j + qual_ceil_radius ); u++ )
            {    newvals[j] = Min( newvals[j], (int) (*quals)[id][u] );    }    }
        for ( longlong j = 0; j < (*quals)[id].size( ); j++ )
          (*quals)[id].set( j, newvals[j] );    }

}
