/////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                   //
//       This software and its documentation are copyright (2008) by the   //
//   Broad Institute/Massachusetts Institute of Technology.  All rights    //
//   are reserved.  This software is supplied without any warranty or      //
//   guaranteed support whatsoever. Neither the Broad Institute nor MIT    //
//   can be responsible for its use, misuse, or functionality.             //
/////////////////////////////////////////////////////////////////////////////
  
/*
 * StdioWrapper.h
 * stdio calls with error checking
 */

#ifndef STDIOWRAPPER_H_
#define STDIOWRAPPER_H_

#include <cerrno>
#include <cstdio>

#include "system/Assert.h"
#include "system/System.h"
#include "CommonSemanticTypes.h"

inline FILE *fopenCheck(const filename_t &f, const char *mode)
{
  FILE *ret = fopen(f.c_str(), mode);
  if (ret == NULL) {
    FatalErr("Error opening file " << f << " errno: " << ErrnoExplanation(errno));
  }
  return ret;
}

inline void fcloseCheck(FILE *fp) 
{
  if (fclose(fp) != 0) {
    FatalErr("Error closing file: " << ErrnoExplanation(errno));
  }
}

inline void fwriteCheck(const void *ptr, size_t size, size_t nmemb, FILE *stream)
{
  size_t ret = fwrite(ptr, size, nmemb, stream);
  if (ret != nmemb) {
    FatalErr("Error writing file: " << ErrnoExplanation(errno));
  }
}

void freadCheck(void *ptr, size_t size, size_t nmemb, FILE *stream);

/**
 * fseek return value doesn't make much sense, so don't return it. 
 */
inline void fseekCheck(FILE *stream, long offset, int whence)
{
  int ret = fseek(stream, offset, whence);
  if (ret == -1) {
    FatalErr("Error fseeking: " << ErrnoExplanation(errno));
  }
}

inline long ftellCheck(FILE *stream)
{
  long ret = ftell(stream);
  if (ret == -1) {
    FatalErr("Error ftelling: " << ErrnoExplanation(errno));
  }
  return ret;
}
#endif /* STDIOWRAPPER_H_ */
