///////////////////////////////////////////////////////////////////////////////
//                   SOFTWARE COPYRIGHT NOTICE AGREEMENT                     //
//       This software and its documentation are copyright (2010) by the     //
//   Broad Institute.  All rights are reserved.  This software is supplied   //
//   without any warranty or guaranteed support whatsoever. The Broad        //
//   Institute is not responsible for its use, misuse, or functionality.     //
///////////////////////////////////////////////////////////////////////////////
/*
 * \file SysConf.cc
 * \author tsharpe
 * \date Apr 15, 2010
 *
 * \brief
 */
#ifndef FORCE_DEBUG
#define NDEBUG
#endif

#include "system/SysConf.h"
#include "system/Exit.h"
#include <iostream>

namespace
{

void sysconfErr( char const* attr )
{
    std::cerr << "Sysconf unable to determine value of " << attr << std::endl;
    CRD::exit(1);
}

size_t gPagSiz;
size_t gPhysPages;
size_t gProcsOnline;
size_t gClockTicksPerSec;

}

size_t pageSize()
{
    if ( !gPagSiz )
    {
        long result = sysconf(_SC_PAGESIZE);
        if ( result == -1 ) sysconfErr("_SC_PAGESIZE");
        gPagSiz = result;
    }
    return gPagSiz;
}

size_t physicalMemory()
{
    if ( !gPhysPages )
    {
        long result = sysconf(_SC_PHYS_PAGES);
        if ( result == -1 ) sysconfErr("_SC_PHYS_PAGES");
        gPhysPages = result;
    }
    return gPhysPages*pageSize();
}

size_t processorsOnline()
{
    if ( !gProcsOnline )
    {
        long result = sysconf(_SC_NPROCESSORS_ONLN);
        if ( result == -1 ) sysconfErr("_SC_NPROCESSORS_ONLN");
        gProcsOnline = result;
    }
    return gProcsOnline;
}

size_t clockTicksPerSecond()
{
    if ( !gClockTicksPerSec )
    {
        long result = sysconf(_SC_CLK_TCK);
        if ( result == -1 ) sysconfErr("_SC_CLK_TCK");
        gClockTicksPerSec = result;
    }
    return gClockTicksPerSec;
}
